@@ -44,17 +44,20 @@ extern "C" {
 #include <unistd.h>
 #endif
 
-#include "meta_modelica.h"
+#include "meta/meta_modelica.h"
 #include <limits.h>
 #include "ModelicaUtilities.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#if defined(__MINGW32__)
+#define _POSIX_THREAD_SAFE_FUNCTIONS 200112L /* for ctime_r in time.h */
+#endif
 #include <time.h>
 #include <math.h>
 
-#include "rtclock.h"
+#include "util/rtclock.h"
 #include "omc_config.h"
 #include "errorext.h"
 #include "settingsimpl.h"
@@ -76,6 +79,7 @@ typedef void* iconv_t;
 
 #if defined(__MINGW32__) || defined(_MSC_VER)
 #include <rpc.h>
+#include <psapi.h>
 #define getFunctionPointerFromDLL  GetProcAddress
 #define FreeLibraryFromHandle !FreeLibrary
 
@@ -302,7 +306,7 @@ extern char* SystemImpl__pwd(void)
     return NULL;
   }
 
-  WCHAR unicodePath[bufLen];
+  WCHAR* unicodePath = (WCHAR*)omc_alloc_interface.malloc_atomic(bufLen * sizeof(WCHAR));
   if (!GetCurrentDirectoryW(bufLen, unicodePath)) {
     c_add_message(NULL,-1,ErrorType_scripting,ErrorLevel_error,gettext("GetCurrentDirectoryW failed."),NULL,0);
     return NULL;
@@ -312,6 +316,7 @@ extern char* SystemImpl__pwd(void)
   SystemImpl__toWindowsSeperators(buffer, bufferLength);
   char *res = omc_alloc_interface.malloc_strdup(buffer);
   MULTIBYTE_OR_WIDECHAR_VAR_FREE(buffer);
+  GC_free(unicodePath);
   return res;
 #else
   char buf[MAXPATHLEN];
@@ -421,7 +426,7 @@ static char* SystemImpl__readFile(const char* filename)
   }
   buf = (char*) omc_alloc_interface.malloc_atomic(statstr.st_size+1);
 
-  if( (res = fread(buf, sizeof(char), statstr.st_size, file)) != statstr.st_size) {
+  if( (res = omc_fread(buf, sizeof(char), statstr.st_size, file, 0)) != statstr.st_size) {
     const char *c_tokens[2]={strerror(errno),filename};
     c_add_message(NULL,85, /* ERROR_OPENING_FILE */
       ErrorType_scripting,
@@ -630,8 +635,8 @@ int SystemImpl__systemCall(const char* str, const char* outFile)
   fflush(NULL); /* flush output so the testsuite is deterministic */
 #if defined(__MINGW32__) || defined(_MSC_VER)
   if (*outFile) {
-    char *command = (char *)omc_alloc_interface.malloc_atomic(strlen(str) + strlen(outFile) + 10);
-    sprintf(command, "%s >> %s 2>&1", str, outFile);
+    char *command = (char *)omc_alloc_interface.malloc_atomic(strlen(str) + strlen(outFile) + 12);
+    sprintf(command, "%s >> \"%s\" 2>&1", str, outFile);
     status = runProcess(command);
     GC_free((void*)command);
   } else {
@@ -870,11 +875,7 @@ int SystemImpl__spawnCall(const char* path, const char* str)
 
 int SystemImpl__plotCallBackDefined(threadData_t *threadData)
 {
-  if (threadData->plotClassPointer && threadData->plotCB) {
-    return 1;
-  } else {
-    return 0;
-  }
+  return threadData->plotClassPointer && threadData->plotCB;
 }
 
 void SystemImpl__plotCallBack(threadData_t *threadData, int externalWindow, const char* filename, const char* title, const char* grid, const char* plotType,
@@ -882,13 +883,26 @@ void SystemImpl__plotCallBack(threadData_t *threadData, int externalWindow, cons
                               const char* y2, const char* curveWidth, const char* curveStyle, const char* legendPosition, const char* footer, const char* autoScale,
                               const char* variables)
 {
-  if (threadData->plotClassPointer && threadData->plotCB) {
+  if (SystemImpl__plotCallBackDefined(threadData)) {
     PlotCallback pcb = threadData->plotCB;
     pcb(threadData->plotClassPointer, externalWindow, filename, title, grid, plotType, logX, logY, xLabel, yLabel, x1, x2, y1, y2, curveWidth, curveStyle,
         legendPosition, footer, autoScale, variables);
   }
 }
 
+int SystemImpl__loadModelCallBackDefined(threadData_t *threadData)
+{
+  return threadData->loadModelClassPointer && threadData->loadModelCB;
+}
+
+void SystemImpl__loadModelCallBack(threadData_t *threadData, const char* modelname)
+{
+  if (SystemImpl__loadModelCallBackDefined(threadData)) {
+    LoadModelCallback cb = threadData->loadModelCB;
+    cb(threadData->loadModelClassPointer, modelname);
+  }
+}
+
 extern double SystemImpl__time(void)
 {
   clock_t cl = clock();
@@ -905,7 +919,11 @@ extern int SystemImpl__directoryExists(const char *str)
   char* path = strdup(str);
   int last = strlen(path)-1;
   /* adrpo: RTFM! the path cannot end in a slash??!! https://msdn.microsoft.com/en-us/library/windows/desktop/aa364418(v=vs.85).aspx */
-  if (last > 0 && (path[last] == '\\' || path[last] == '/')) path[last] = '\0';
+  while (last > 0 && (path[last] == '\\' || path[last] == '/'))
+      last--;
+
+  path[last + 1] = '\0';
+
   sh = FindFirstFile(path, &FileData);
   free(path);
   if (sh == INVALID_HANDLE_VALUE)
@@ -970,7 +988,7 @@ extern int SystemImpl__copyFile(const char *str_1, const char *str_2)
     return 0;
   }
 
-  while ( n = fread(buf, 1, 8192, source) ) {
+  while (( n = omc_fread(buf, 1, 8192, source, 1) )) {
     if (n != fwrite(buf, 1, n, target)) {
       rv = 0;
       break;
@@ -1242,7 +1260,7 @@ extern const char* SystemImpl__readFileNoNumeric(const char* filename)
   file = omc_fopen(filename,"rb");
   buf = (char*) omc_alloc_interface.malloc_atomic(statstr.st_size+1);
   bufRes = (char*) omc_alloc_interface.malloc_atomic((statstr.st_size+70)*sizeof(char));
-  if( (res = fread(buf, sizeof(char), statstr.st_size, file)) != statstr.st_size) {
+  if( (res = omc_fread(buf, sizeof(char), statstr.st_size, file, 0)) != statstr.st_size) {
     fclose(file);
     return "Failed while reading file";
   }
@@ -1322,8 +1340,8 @@ static const char* SystemImpl__getUUIDStr(void)
 {
   static char uuidStr[37] = "8c4e810f-3df3-4a00-8276-176fa3c9f9e0";
 #if defined(__MINGW32__) || defined(_MSC_VER)
-  unsigned char *tmp;
-  UUID uuid;
+  unsigned char *tmp = NULL;
+  UUID uuid = {0};
   if (UuidCreate(&uuid) == RPC_S_OK)
     UuidToString(&uuid, &tmp);
   tmp[36] = '\0';
@@ -1340,7 +1358,7 @@ static const char* SystemImpl__getUUIDStr(void)
 typedef void (*mmc_GC_function_set_gc_state)(mmc_GC_state_type*);
 
 #if defined(__MINGW32__) || defined(_MSC_VER)
-int SystemImpl__loadLibrary(const char *str, int printDebug)
+int SystemImpl__loadLibrary(const char *str, int relativePath, int printDebug)
 {
   char libname[MAXPATHLEN];
   char currentDirectory[MAXPATHLEN];
@@ -1350,15 +1368,31 @@ int SystemImpl__loadLibrary(const char *str, int printDebug)
   HMODULE h;
   const char* ctokens[2];
   mmc_GC_function_set_gc_state mmc_GC_set_state_lib_function = NULL;
-  /* adrpo: use BACKSLASH here as specified here: http://msdn.microsoft.com/en-us/library/ms684175(VS.85).aspx */
-  GetCurrentDirectory(bufLen,currentDirectory);
-#if defined(_MSC_VER)
-  _snprintf(libname, MAXPATHLEN, "%s\\%s.dll", currentDirectory, str);
-#else
-  snprintf(libname, MAXPATHLEN, "%s\\%s.dll", currentDirectory, str);
-#endif
 
-  h = LoadLibrary(libname);
+  if (str[0] != '\0') {
+    /* adrpo: use BACKSLASH here as specified here: http://msdn.microsoft.com/en-us/library/ms684175(VS.85).aspx */
+    if (relativePath) {
+      GetCurrentDirectory(bufLen,currentDirectory);
+#if defined(_MSC_VER)
+      _snprintf(libname, MAXPATHLEN, "%s\\%s", currentDirectory, str);
+#else
+      snprintf(libname, MAXPATHLEN, "%s\\%s", currentDirectory, str);
+#endif
+    } else {
+#if defined(_MSC_VER)
+      _snprintf(libname, MAXPATHLEN, "%s", str);
+#else
+      snprintf(libname, MAXPATHLEN, "%s", str);
+#endif
+    }
+
+    h = LoadLibrary(libname);
+  } else { /* no library name, fetch current module handle */
+    /* set the libname */
+    strcpy(libname, "[own process]");
+    h = GetModuleHandle(0);
+  }
+
   if (h == NULL) {
     LPVOID lpMsgBuf;
     FormatMessage(
@@ -1377,16 +1411,6 @@ int SystemImpl__loadLibrary(const char *str, int printDebug)
     return -1;
   }
 
-  /* adrpo, pass the mmc_GC_state pointer from the current process!
-  mmc_GC_set_state_lib_function = (mmc_GC_function_set_gc_state)getFunctionPointerFromDLL(h, "mmc_GC_set_state");
-  if (mmc_GC_set_state_lib_function == NULL) {
-    fprintf(stderr, "Unable to get pointer for mmc_GC_set_state in  %s!\n", libname);
-    fflush(stderr);
-    return -1;
-  }
-  mmc_GC_set_state_lib_function(mmc_GC_state);
-  */
-
   libIndex = alloc_ptr();
   if (libIndex < 0) {
     //fprintf(stderr, "Error loading library %s!\n", libname); fflush(stderr);
@@ -1401,7 +1425,7 @@ int SystemImpl__loadLibrary(const char *str, int printDebug)
 }
 
 #else
-int SystemImpl__loadLibrary(const char *str, int printDebug)
+int SystemImpl__loadLibrary(const char *str, int relativePath, int printDebug)
 {
   char libname[MAXPATHLEN];
   modelica_ptr_t lib = NULL;
@@ -1409,12 +1433,23 @@ int SystemImpl__loadLibrary(const char *str, int printDebug)
   void *h = NULL;
   mmc_GC_function_set_gc_state mmc_GC_set_state_lib_function = NULL;
   const char* ctokens[2];
-  snprintf(libname, MAXPATHLEN, "./%s" CONFIG_DLL_EXT, str);
 #if defined(RTLD_DEEPBIND)
-  h = dlopen(libname, RTLD_LOCAL | RTLD_NOW | RTLD_DEEPBIND);
+  int flags = RTLD_LOCAL | RTLD_NOW | RTLD_DEEPBIND;
 #else
-  h = dlopen(libname, RTLD_LOCAL | RTLD_NOW);
+  int flags = RTLD_LOCAL | RTLD_NOW;
 #endif
+
+  if (str[0] != '\0') {
+    if (relativePath) {
+      snprintf(libname, MAXPATHLEN, "./%s", str);
+    } else {
+      snprintf(libname, MAXPATHLEN, "%s", str);
+    }
+    h = dlopen(libname, flags);
+  } else {
+    h = dlopen(NULL, flags);
+  }
+
   if (h == NULL) {
     ctokens[0] = dlerror();
     ctokens[1] = libname;
@@ -1422,16 +1457,6 @@ int SystemImpl__loadLibrary(const char *str, int printDebug)
     return -1;
   }
 
-  /* adrpo, pass the mmc_GC_state pointer from the current process!
-  mmc_GC_set_state_lib_function = (mmc_GC_function_set_gc_state)getFunctionPointerFromDLL(h, "mmc_GC_set_state");
-  if (mmc_GC_set_state_lib_function == NULL) {
-    fprintf(stderr, "Unable to get pointer for mmc_GC_set_state in  %s!\n", libname);
-    fflush(stderr);
-    return -1;
-  }
-  mmc_GC_set_state_lib_function(mmc_GC_state);
-  */
-
   libIndex = alloc_ptr();
   if (libIndex < 0) {
     fprintf(stderr, "Error loading library %s!\n", libname); fflush(stderr);
@@ -1507,6 +1532,7 @@ int SystemImpl__lookupFunction(int libIndex, const char *str)
   modelica_ptr_t lib = NULL, func = NULL;
   function_t funcptr;
   int funcIndex;
+  long lastError = 0;
 
   lib = lookup_ptr(libIndex);
 
@@ -1514,9 +1540,58 @@ int SystemImpl__lookupFunction(int libIndex, const char *str)
     return -1;
 
   funcptr =  (int (*)(threadData_t*, type_description*, type_description*)) getFunctionPointerFromDLL(lib->data.lib, str);
+  lastError = GetLastError();
+
+#if defined(__MINGW32__) || defined(_MSC_VER)
+  /* windows is special :)
+   * if we didn't find the function, search it in all the loaded DLLs as Linux does
+   */
+  if (funcptr == NULL) {
+    HMODULE hMods[1024];
+    HMODULE hProcess = GetCurrentProcess();
+    DWORD cbNeeded;
+    unsigned int i;
+    char szModName[MAX_PATH];
+
+    /* see if the GetModuleHandle(0) is in the current process and an executable */
+    if ( GetModuleFileNameEx( hProcess, lib->data.lib, szModName, sizeof(szModName) / sizeof(char)) )
+    {
+      // if strcmp(szModName[strlen(szModName)-4], ".exe")
+      if ( EnumProcessModules(hProcess, hMods, sizeof(hMods), &cbNeeded) )
+      {
+        for ( i = 0; i < (cbNeeded / sizeof(modelica_integer)); i++ )
+        {
+          funcptr = (int (*)(threadData_t*, type_description*, type_description*)) getFunctionPointerFromDLL(hMods[i], str);
+          if (funcptr != NULL)
+          {
+            break;
+          }
+
+          /* // uncomment for debugging
+          {
+            if ( GetModuleFileNameEx( hProcess, hMods[i], szModName, sizeof(szModName) / sizeof(char)) )
+            {
+              fprintf(stderr, "function %s in %s handle:[%p][%d] funcptr[%p] GetLastError:%d\n", str, szModName, hMods[i], i, funcptr, GetLastError()); fflush(NULL);
+            }
+          }
+          */
+        }
+      }
+    }
+  }
+#endif
 
   if (funcptr == NULL) {
-    fprintf(stderr, "Unable to find `%s': %lu.\n", str, GetLastError());
+    const char* err_toks[2];
+    char id_buf[11];
+    snprintf(id_buf, 11, "%lu", lastError);
+#if defined(__MINGW32__) || defined(_MSC_VER)
+    err_toks[0] = id_buf;
+#else
+    err_toks[0] = dlerror();
+#endif
+    err_toks[1] = str;
+    c_add_message(NULL, -1, ErrorType_runtime, ErrorLevel_error, gettext("Unable to find `%s': %s.\n"), err_toks, 2);
     return -1;
   }
 
@@ -1720,40 +1795,77 @@ extern char* SystemImpl__unescapedString(const char* str)
   return res;
 }
 
-extern char* System_forceQuotedIdentifier(const char* str)
-{
-  const char lookupTbl[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
-  char *res,*cur;
-  int len,i;
-  const int offset = 10;
-  const char _omcQuot[]="_omcQuot_";
-  len = strlen(str);
-  res = (char*) omc_alloc_interface.malloc_atomic(2*len+offset+64);
-  cur = res;
-  cur += sprintf(cur,"%s",_omcQuot);
-  for (i=0; i<len; i++) {
-    unsigned char c = str[i];
-    *cur = lookupTbl[c/16];
-    cur++;
-    *cur = lookupTbl[c%16];
-    cur++;
+
+static int compute_sanitized_string_size(const char* str) {
+  int i, count;
+
+  for (i=0, count=0; str[i]; i++, count++) {
+    // Each non-alphanum character needs two more char for its
+    // escape + two hex representation
+    if (!isalnum(str[i])) {
+      count += 2;
+    }
   }
+
+  return count;
+}
+
+static char* sanitize_string(const char* src, char* dst) {
+  const char lookupTbl[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
+
+  int i;
+  while (*src) {
+    unsigned char c = *src++;
+    if (isalnum(c)) {
+      *dst = c;
+      dst++;
+    } else {
+      *dst++ = '_';
+      *dst++ = lookupTbl[c/16];
+      *dst++ = lookupTbl[c%16];
+    }
+  }
+
+  return dst;
+}
+
+// This function does not assume the input is actually a quoted string e.g 'gb' or 'ab[!%'
+//
+extern char* System_sanitizeQuotedIdentifier(const char* str)
+{
+  char *res,*cur;
+
+  const char openquote[]="_omcQ";
+  const int qsize = sizeof(openquote) - 1;
+
+  // Each non-alphanum character needs one more char for its
+  // two char ascii representation.
+  int nrchars_needed = compute_sanitized_string_size(str) + qsize;
+
+  res = (char*) omc_alloc_interface.malloc_atomic((nrchars_needed+1) * sizeof(char));
+
+  cur = res;
+  cur += sprintf(cur, "%s", openquote);
+  cur = sanitize_string(str, cur);
   *cur = '\0';
+
+  assert((cur == res + nrchars_needed) && "Allocated memory does not exactly fit the unquoted string output");
+
   return res;
 }
 
-extern char* SystemImpl__unquoteIdentifier(const char* str)
+extern char* SystemImpl__unquoteIdentifier(char* str)
 {
-  char c = *str;
-  if ((c == '\'')
-#if !defined(OPENMODELICA_BOOTSTRAPPING_STAGE_1)
-   || strstr(str, "$")
-#endif
-  )
-  {
-    return System_forceQuotedIdentifier(str);
+
+  if (str[0] == '\'') {
+    return System_sanitizeQuotedIdentifier(str);
   }
-  return NULL;
+
+  if (strstr(str, "$")) {
+    return System_sanitizeQuotedIdentifier(str);
+  }
+
+  return str;
 }
 
 #define TIMER_MAX_STACK  1000
@@ -1873,18 +1985,14 @@ static int SystemImpl__uriToClassAndPath(const char *uri, const char **scheme, c
   return 1;
 }
 
-#ifdef NO_LAPACK
-int SystemImpl__dgesv(void *lA, void *lB, void **res)
-{
-  MMC_THROW();
-}
-#else
 /* adrpo 2011-06-23
  * extern definition to dgesv_ from -llapack
  * as we do not link with -lsim and the one
  * in matrix.h got renamed to _omc_dgesv_ to
  * avoid name clashes!
  */
+#ifdef HAVE_LAPACK
+
 extern int dgesv_(integer *n, integer *nrhs, doublereal *a, integer *lda, integer *ipiv, doublereal *b, integer *ldb, integer *info);
 
 int SystemImpl__dgesv(void *lA, void *lB, void **res)
@@ -1926,6 +2034,14 @@ int SystemImpl__dgesv(void *lA, void *lB, void **res)
   *res = tmp;
   return info;
 }
+
+#else
+int SystemImpl__dgesv(void *lA, void *lB, void **res)
+{
+  c_add_message(NULL, -1, ErrorType_runtime, ErrorLevel_error,
+                "A LAPACK routine is called but OMC is not compiled with LAPACK support", NULL, 0);
+  MMC_THROW();
+}
 #endif
 
 #ifdef NO_LPLIB
@@ -1992,13 +2108,17 @@ typedef struct {
   int fileIsDir;
 } modelicaPathEntry;
 
-void splitVersion(const char *version, long *versionNum, char **versionExtra)
+int splitVersion(const char *version, long *versionNum, char **versionExtra)
 {
   const char *buf = version;
   char *next;
   long l;
   int cont,i=0,len;
   memset(versionNum,0,sizeof(long)*MODELICAPATH_LEVELS);
+  if (!isdigit(version[0])) {
+    *versionExtra = omc_alloc_interface.malloc_strdup(buf);
+    return 0;
+  }
   do {
     /* fprintf(stderr, "look versionNum %s\n", buf); */
     l = strtol(buf,&next,10);
@@ -2017,6 +2137,7 @@ void splitVersion(const char *version, long *versionNum, char **versionExtra)
   if (len >= 2 && 0==strcmp("mo", *versionExtra+len-2)) {
     (*versionExtra)[len-2] = '\0';
   }
+  return 1;
 }
 
 static int regularFileExistsInDirectory(const char *dir1, const char *dir2, const char *file)
@@ -2132,12 +2253,13 @@ static int modelicaPathEntryVersionGreater(long *ver1, long *ver2, int numToTest
 
 static int getLoadModelPathFromSingleTarget(const char *searchTarget, modelicaPathEntry *entries, int numEntries, int exactVersion, const char **outDir, char **outName, int *isDir)
 {
-  int i, j, foundIndex = -1;
+  int i, j, foundIndex = -1, foundDigitVersion;
   long version[MODELICAPATH_LEVELS] = {0}, foundVersion[MODELICAPATH_LEVELS] = {0};
   char *versionExtra;
-  splitVersion(searchTarget,version,&versionExtra);
+  /* fprintf(stderr, "searchTarget %s\n", searchTarget); */
+  foundDigitVersion = splitVersion(searchTarget,version,&versionExtra);
   /* fprintf(stderr, "expected %ld.%ld.%ld.%ld %s ; exact=%d\n", version[0], version[1], version[2], version[3], versionExtra, exactVersion); */
-  if (version > 0 && !*versionExtra) {
+  if (foundDigitVersion && !*versionExtra) {
     /* Makes us load 3.2.1 if 3.2.0.0 is not available.
      * Note that all 4 levels are present and 3.2 is equivalent to 3.2.0.0
      * Search one additional level for each time we fail.
@@ -2149,7 +2271,7 @@ static int getLoadModelPathFromSingleTarget(const char *searchTarget, modelicaPa
 
         if (modelicaPathEntryVersionEqual(entries[i].version,version,j)
             && (j==MODELICAPATH_LEVELS || modelicaPathEntryVersionGreater(entries[i].version,version,MODELICAPATH_LEVELS))
-            && entries[i].versionExtra[0] == '\0') {
+            && (entries[i].versionExtra[0] == '\0' || entries[i].versionExtra[0] == '+' || entries[i].versionExtra[0] == '-')) {
           if (modelicaPathEntryVersionGreater(entries[i].version,foundVersion,MODELICAPATH_LEVELS)) {
             memcpy(foundVersion,entries[i].version,sizeof(long)*MODELICAPATH_LEVELS);
             foundIndex = i;
@@ -2164,18 +2286,28 @@ static int getLoadModelPathFromSingleTarget(const char *searchTarget, modelicaPa
       }
     }
   }
-  if (*versionExtra) {
-    /* fprintf(stderr, "Look for version %lx versionExtra: %s\n", version, versionExtra); */
-    for (i=0; i<numEntries; i++) {
-      /* fprintf(stderr, "entry %s/%s\n", entries[i].dir, entries[i].file);
-      fprintf(stderr, "is %ld.%ld.%ld.%ld %s\n", entries[i].version[0], entries[i].version[1], entries[i].version[2], entries[i].version[3], entries[i].versionExtra); */
-      if (modelicaPathEntryVersionEqual(entries[i].version,version,MODELICAPATH_LEVELS) && 0==strncmp(entries[i].versionExtra,versionExtra,strlen(versionExtra))) {
+  /* fprintf(stderr, "Look for version %lx versionExtra: %s\n", version, versionExtra); */
+  for (i=0; i<numEntries; i++) {
+    /* fprintf(stderr, "entry %s/%s\n", entries[i].dir, entries[i].file);
+    fprintf(stderr, "is %ld.%ld.%ld.%ld %s\n", entries[i].version[0], entries[i].version[1], entries[i].version[2], entries[i].version[3], entries[i].versionExtra); */
+    if (version[0]==0 && version[1]==0 && version[2]==0) {
+      const char *entryVersionExtra = entries[i].versionExtra;
+      if (entryVersionExtra[0] == '-' && versionExtra[0] != '-') {
+        entryVersionExtra = entryVersionExtra+1;
+      }
+      if (0==strncmp(entryVersionExtra,versionExtra,strlen(versionExtra))) {
         *outDir = entries[i].dir;
         *outName = entries[i].file;
         *isDir = entries[i].fileIsDir;
         return 0;
       }
     }
+    if (modelicaPathEntryVersionEqual(entries[i].version,version,MODELICAPATH_LEVELS) && 0==strncmp(entries[i].versionExtra,versionExtra,strlen(versionExtra))) {
+      *outDir = entries[i].dir;
+      *outName = entries[i].file;
+      *isDir = entries[i].fileIsDir;
+      return 0;
+    }
   }
   return 1;
 }
@@ -2188,7 +2320,7 @@ static int getLoadModelPathFromDefaultTarget(const char *name, modelicaPathEntry
 
   /* Look for best release version */
   for (i=0; i<numEntries; i++) {
-    if (modelicaPathEntryVersionGreater(entries[i].version,foundVersion,MODELICAPATH_LEVELS) && entries[i].versionExtra[0] == '\0') {
+    if (modelicaPathEntryVersionGreater(entries[i].version,foundVersion,MODELICAPATH_LEVELS) && (entries[i].versionExtra[0] == '\0' || entries[i].versionExtra[0] == '+' || entries[i].versionExtra[0] == '-')) {
       memcpy(foundVersion,entries[i].version,sizeof(long)*MODELICAPATH_LEVELS);
       foundIndex = i;
     }
@@ -2329,14 +2461,6 @@ extern int SystemImpl_tmpTickMaximum(threadData_t *threadData, int index)
   return data->tmp_tick_max_no[index];
 }
 
-#if defined(OPENMODELICA_BOOTSTRAPPING_STAGE_1)
-extern int SystemImpl_tmpTick(threadData_t *threadData)
-{
-  int res = SystemImpl_tmpTickIndex(threadData,0);
-  return res;
-}
-#endif
-
 extern void SystemImpl_tmpTickReset(threadData_t *threadData, int start)
 {
   SystemImpl_tmpTickResetIndex(threadData,start,0);
@@ -2454,7 +2578,7 @@ extern const char* SystemImpl__iconv(const char * str, const char *from, const c
   return buf;
 }
 
-#include <tinymt64.h>
+#include "util/tinymt64.h"
 
 static tinymt64_t system_random_seed = {{0,0},0,0,0};
 
@@ -2596,7 +2720,7 @@ int System_getTerminalWidth(void)
 #endif
 }
 
-#include "simulation_options.h"
+#include "util/simulation_options.h"
 
 #define SB_SIZE 8192*4
 #define SB_SIZE_MINUS_ONE (SB_SIZE-1)
@@ -2820,8 +2944,8 @@ int SystemImpl__fileContentsEqual(const char *file1, const char *file2)
     return 0;
   }
   do {
-    i1 = fread(buf1,1,8192,f1);
-    i2 = fread(buf2,1,8192,f2);
+    i1 = omc_fread(buf1,1,8192,f1, 1);
+    i2 = omc_fread(buf2,1,8192,f2, 1);
     if (i1 != i2 || strncmp(buf1,buf2,i1)) {
       error = 1;
     }
@@ -2879,16 +3003,34 @@ static char *omc_strtok_r(char *str, const char *delim, char **saveptr)
 
 #endif /* defined(__MINGW32__) */
 
-int SystemImpl__stat(const char *filename, double *size, double *mtime)
+typedef enum {
+    FileType_NoFile = 1,
+    FileType_RegularFile,
+    FileType_Directory,
+    FileType_SpecialFile
+} OpenModelicaFileType;
+
+int SystemImpl__stat(const char *filename, double *size, double *mtime, int *fileType)
 {
   struct stat stats;
   if (0 != stat(filename, &stats)) {
     *size = 0;
     *mtime = 0;
+    *fileType = FileType_NoFile;
     return 0;
   }
   *size = stats.st_size;
   *mtime = stats.st_mtime;
+  if (S_ISREG(stats.st_mode)) {
+    *fileType = FileType_RegularFile;
+  }
+  else if (S_ISDIR(stats.st_mode)) {
+    *fileType = FileType_Directory;
+  }
+  else {
+    *fileType = FileType_SpecialFile;
+  }
+
   return 1;
 }
 
@@ -2958,7 +3100,7 @@ int SystemImpl__covertTextFileToCLiteral(const char *textFile, const char *outFi
     fputc('{', fout);
     fputc('\n', fout);
     do {
-      n = fread(buffer,1,511,fin);
+      n = omc_fread(buffer,1,511,fin, 1);
       j = 0;
       /* adrpo: encode each char */
       for (i=0; i<n; i++) {
@@ -3001,7 +3143,7 @@ int SystemImpl__covertTextFileToCLiteral(const char *textFile, const char *outFi
   {
     fputc('\"', fout);
     do {
-      n = fread(buffer,1,511,fin);
+      n = omc_fread(buffer,1,511,fin, 1);
       j = 0;
       for (i=0; i<n; i++) {
         switch (buffer[i]) {
